#! /usr/bin/env python2

# SSGridView.py

"""spreadsheet grid using a.o. the Tkinter Canvas widget

Demonstrates the creation of widgets as part of a Canvas object
This is used to build a simple spreadsheet GUI
"""
__author__ =  'msdl.cs.mcgill.ca'
__version__=  '0.4'

# revisions:
#
#    24/03/2004 Marc Provost
#     added support for cell editing
#     added support for named colums
#     moved all behaviour code to a statechart
#
#    10/01/2004 Hans Vangheluwe
#     refactored
#
#    04/12/2002 Jean-Sebastien Bolduc
#     thoroughly cleaned up the prototype
#
#    01/03/2002 Hans Vangheluwe
#     the initial prototype

# Required modules

from   Tkinter      import *  # basic Tkinter
import tkFileDialog           # file selection (to output Postscript in)
import tkMessageBox

import types                  # for type checking

from   spreadsheet.SSheetErrors import *
from   SSGridViewStateChart import SSGridViewStateChart # the behaviour


def stringToInt(col):
    """<string> -> <int>

    Info: Function used to transform strings used as
          cell column references (e.g., AA in  AA25) by integers
          for efficiency reasons.  Note that 0 are removed
          from possible integers to avoid ambiguity.
          (i.e., if "A" is converted to 0, "AAA"=000=0,
           0==0 while "A" != "AAA").
    Complexity: O(length(string)) (in our case len(col) <= 3)
    """
    base=ord("A")
    answer=0
    for l in xrange(len(col)):
        answer+=(ord(col[-(l+1)])-base+1)*int(pow(26,l))
    return answer

def intToString(integer):
    """<int> -> <string>

    Info: Inverse of stringToInt()
    Complexity: O(1) since integer assumed <= 26^3
    """
    ans=""
    base=ord("A")
    i=2
    for i in [2,1,0]:
        tmp=int(pow(26,i))
        tmpInteger=integer
        for j in xrange(i):
            tmpInteger-=int(pow(26,j))
        if tmp<=tmpInteger or i==0:
            charValue=(tmpInteger-tmpInteger%tmp)/tmp
            integer=integer-charValue*tmp
            char=chr(charValue+base-1)
            ans+=char
    return ans

class InvalidKeyError(Exception):
  """
  Exception indicating a non-acceptable key was pressed
  """
  pass

class CellCoord:
  """
  A coordinate class (unbounded)
  To enhance readability:
  refer to coord.row and coord.col
  rather than coord[0] and coord[1]
  """

  def __init__(self, row=0, col=0):
    self.row=row
    self.col=col

  def __repr__(self):
    """string representation
    """
    return intToString(self.col+1)+str(self.row+1)


class SpreadsheetGridView_Static(Frame):
  """The Tkinter frame containing a GridView observing a
  subject, a PolyCell instance.
  """

  def __init__(self,
               ID,
               subject=None,
               master=None,
               title="Spreadsheet View",
               dataOffsetHor=0, dataOffsetVer=0,
               height=30, width=7,
               cellHeight=18, cellWidth=50,
               version="V"+__version__):
    """Constructor

    Arguments:
     ID            : the unique ID of this particular SpreadsheetGridView
     subject       : a PolyCell instance 
     master        : typically a Toplevel window inside which this frame
                     is embedded
     title         : the Toplevel window would have used this to
                     set the window title
     dataOffsetHor : cell row from which to start viewing
     dataOffsetVer : cell column from which to start viewing
     height        : number of rows to view
     width         : number of columns to view
     cellHeight    : the height in pixels of a cell
     cellWidth     : the width in pixels of a cell
     version       : version string
                     default set to the version of SpreadsheetGridView
		     typically replaced by the version of the overall
		     application
    """

    #Initializing the Dynamic part of this GridView.
    #The dynamics of the SpreadsheetGridView is encoded in
    #the class SSGridViewStateChart, which was synthesized
    #by means of the SCC DCharts compiler from a DChart specification.
    #The class SpreadsheetGridView_Dynamic provides a bridge between
    #SpreadsheetGridView_Static and SSGridViewStateChart.
    self.dynamicGridView=SpreadsheetGridView_Dynamic(self)

    #In the rest of the constructor, we construct the
    #initial default GridView that will be dynamically
    #modified by the statechart.

    # the ID of this static gridview
    self.__ID=ID

    # Make arguments available to all methods of this class
    # Note: should really test for validity
    #       for example, height and width should be >= 1

    # the window's title
    self.__title = title

    # height in pixels of one cell
    self.__cellHeight = cellHeight

    # width in pixels of one cell
    self.__cellWidth  = cellWidth

    # number of rows
    self.__height=height

    # number of columns
    self.__width=width


    # X position of each column (each column is a canvas line)
    self.__cols=[]

    # TK tag mapped to each column (each column is a canvas line)
    self.__colsTag=[]

    # TK tag mapped to the canvas column text (i.e. A B C D ...)
    self.__colsTagIndex=[]

    # Y position of each row (each row is a canvas line)
    self.__rows=[]

    # TK tag mapped to each row (each column is a canvas line)
    self.__rowsTag=[]

    # TK tag mapped to the canvas row text (i.e. 1 2 3 4 ... )
    self.__rowsTagIndex=[]

    self.__version=version

    # Setting up the cols X position array
    deltaX=cellWidth
    for col in xrange(width+1):
      self.__cols.append(deltaX)
      deltaX+=cellWidth

    # Setting up the rows Y position array
    deltaY=cellHeight
    for row in xrange(height+1):
      self.__rows.append(deltaY)
      deltaY+=cellHeight

    # Initialize superclass
    Frame.__init__(self, master)

    # parent frame 
    self.root=master
    # what to do when the window is closed 
    # (event received from the Window Manager)
    self.root.protocol("WM_DELETE_WINDOW", self.wmQuit)

    # Datastructure holding the cell formulas.  In this version,
    # The observer pattern is not implemented, so we have
    # one observer per subject.
    # Here, each observer (SpreadsheetGridView) has its own subject
    self.subject=subject

    # Mapping by tuple (col,row) to each text object drawn
    # on the canvas spreadsheet (drawarea)
    self.textOnCanvas={}

    # Initialize packer
    self.pack(fill=BOTH, expand=1)


    # The initial "bounding box" (area filled with data)
    self.maxDim=CellCoord(0,0)

    # Initialize the overCell object
    # which contains the coordinates of the cell
    # the mouse is currently over.
    # This could be any cell, but (0,0) seems most reasonable
    # We make sure this is a meaningful cell (within the
    # spreadsheet) by "snapping" it into the grid boundaries
    # by means of spreadsheetBoundedCoord()
    self.__overCell = self.cellCoords(0,0)

    # The Tk variable overCellString
    # is used to display the coordinates of the cell the mouse is
    # currently over (modulo spreadsheet size)
    # It is updated in the callback routines for events which
    # change the overCell
    self.overCellString = StringVar()
    self.overCellString.set(`self.__overCell`)

    # The Tk variable overCellData
    # is used to display the data in the cell the mouse is currently over
    # (modulo spreadsheet size)
    # It is updated in the callback routine for events which
    # change the overCell
    self.overCellData = StringVar()
    self.overCellData.set("")
    
    # Tk variable mapped to the time counter label
    # Will be updated at every second by the statechart
    # describing the dynamics 
    self.__timeData = StringVar()
    self.__timeData.set("00:00")

    # Initialize selectedCell to be equal to overCell
    # This could be some other cell,
    # but it seems reasonable to make this the same as overCell.
    # Note: selectedCell determines where typed in data will go.
    self.__selectedCell=CellCoord(self.__overCell.row, self.__overCell.col)

    # TK tag mapped to the overcell rectangle
    self.overCellTag=None
    # TK tag mapped to the selectedcell rectangle
    self.selectedCellTag=None

    # Create all widgets in the Frame
    self.__createWidgets()
        
  def __createWidgets(self):
    """ <None> -> <None>

    Function creating the GridView's widgets 
    (at instantiation time).
    """

    self.__createPullDownMenus()

    # Frame to hold top labels and print button
    self.topFrame = Frame(master=self, background='white', border=2)
    self.topFrame.pack(side=TOP,fill=X)

    # Label displaying the content of the cell the mouse is
    # currently over.
    self.overCellContent = Label(master=self.topFrame,
                                 takefocus=0,
                                 textvariable=self.overCellData,
                                 width=20,
                                 relief="groove",
                                 background='white',
                                 foreground='blue')
    self.overCellContent.pack(side=RIGHT, padx=4)

    # Label displaying the cell the mouse is currently over
    # The Tk variable self.overCellString is used for this
    self.overCellCoord = Label(master=self.topFrame,
                               takefocus=0,
                               textvariable=self.overCellString,
                               width=20,
                               relief="groove",
                               background='white',
                               foreground='blue')
    self.overCellCoord.pack(side=RIGHT, padx=4)
    
    # Label displaying the time since the application started
    self.__timeLabel = Label(master=self.topFrame,
                           takefocus=0,
                           textvariable=self.__timeData,
                           width=20,
                           relief="groove",
                           background='white',
                           foreground='black')
    self.__timeLabel.pack(side=RIGHT, padx=4)


    # Canvas (default 15cm x 8cm), stretchable
    # Scrollable region calculated from
    #   self.__height, self.__width
    #   self.__cellHeight, self.__cellWidth
    scrollHor = self.__cols[-1]
    scrollVer = self.__rows[-1]
    self.drawarea = Canvas(master=self,
                           takefocus=1,
                           width="15c", height="8c",
                           background="white",
                           scrollregion=(0, 0, scrollHor, scrollVer))

    # frame for scrollX scrollbar as well as blank corner
    self.scrollXframe = Frame(self)

    # scrollbars for drawarea
    self.scrollX = Scrollbar(self.scrollXframe, takefocus=0, orient=HORIZONTAL)
    self.scrollY = Scrollbar(self, takefocus=0, orient=VERTICAL)

    # link canvas, scrollbars, and scrolling events
    self.drawarea['xscrollcommand'] = self.scrollX.set
    self.drawarea['yscrollcommand'] = self.scrollY.set
    self.scrollX['command'] = self.drawarea.xview
    self.scrollY['command'] = self.drawarea.yview

    # a blank square for aesthetic reasons
    self.blankSquare = Canvas(self.scrollXframe, bg='lightgrey',
                       width=int(self.scrollY['width'])+2,
                       height=int(self.scrollX['width'])+2)

    # pack it all
    self.blankSquare.pack(side=RIGHT)
    self.scrollX.pack(side=LEFT, fill=X, expand=1)
    self.scrollXframe.pack(side=BOTTOM, fill=X)
    self.scrollY.pack(side=RIGHT, fill=Y)

    self.drawarea.pack(fill=BOTH, expand=1)

    # Make sure the Canvas has the focus
    # This is useful if there are other widgets around in the same
    # window such as Entry fields.
    self.drawarea.focus_set()

    self.drawarea.create_rectangle(\
      0,0,self.__cellWidth,self.__cellHeight,
      width=1, outline="grey", fill="#EBF5FF")

    self.__drawColBackground("#EBF5FF")
    self.__drawRowBackground("#EBF5FF")
    self.__drawColIndex()
    self.__drawRowIndex()
    self.__drawRows()
    self.__drawColumns()
    self.__drawOverCell()
    self.__drawSelectedCell()

    # Bind mouse events in the drawarea to callback methods that will
    # generate appropriate events in the behaviour statechart
    self.drawarea.bind("<Double-ButtonPress-1>", self.dynamicGridView.onDrawareaDoubleClick)
    self.drawarea.bind("<Motion>", self.dynamicGridView.onDrawareaMouseMove)
    self.drawarea.bind("<ButtonPress-1>", self.dynamicGridView.onDrawareaMouse1Click)
    self.drawarea.bind("<Left>", self.dynamicGridView.onDrawareaLeft)
    self.drawarea.bind("<Right>", self.dynamicGridView.onDrawareaRight)
    self.drawarea.bind("<Up>", self.dynamicGridView.onDrawareaUp)
    self.drawarea.bind("<Down>", self.dynamicGridView.onDrawareaDown)
    self.drawarea.bind("<Return>", self.dynamicGridView.onDrawareaDown)
    self.drawarea.bind("<KP_Enter>", self.dynamicGridView.onDrawareaDown)
    self.drawarea.bind("<KeyPress>", self.dynamicGridView.onDrawareaKeyPressed)


  def __createPullDownMenus(self):
    """ <None> -> <None>

    Create the menus: File, Edit, etc.
    """

    self.menuBar = Menu(self)

    # create a File pulldown menu, and add it to the menu bar
    self.fileMenu = Menu(self.menuBar, tearoff=0)
    self.fileMenu.add_command(label="New", 
     command=self.subject.mainApp.startPolyCell)
    self.fileMenu.add_command(label="Add gridview...", state="disabled")
    self.fileMenu.add_separator()
    self.fileMenu.add_command(label="Open",state="disabled")
    self.fileMenu.add_command(label="Save",state="disabled")
    self.fileMenu.add_command(label="Save as...",state="disabled")
    self.fileMenu.add_command(label="Print view...", 
     command=self.dynamicGridView.printView)
    self.fileMenu.add_separator()
    self.fileMenu.add_command(label="Exit",command=self.wmQuit)
    self.menuBar.add_cascade(label="File", menu=self.fileMenu)

    # create an Edit pulldown menu, and add it to the menu bar
    self.editMenu = Menu(self.menuBar, tearoff=0)
    self.editMenu.add_command(label="Undo",state="disabled")
    self.editMenu.add_command(label="Redo",state="disabled")
    self.editMenu.add_separator()
    self.editMenu.add_command(label="Cut",state="disabled")
    self.editMenu.add_command(label="Copy",state="disabled")
    self.editMenu.add_command(label="Paste",state="disabled")
    self.editMenu.add_command(label="Paste special",state="disabled")
    self.editMenu.add_separator()
    self.editMenu.add_command(label="Find...",state="disabled")
    self.editMenu.add_command(label="Replace...",state="disabled")
    self.menuBar.add_cascade(label="Edit", menu=self.editMenu)

    self.root.config(menu=self.menuBar)

  def __drawColBackground(self, color="grey"):
    """ <color:String> -> <None>

    Draws the background of the
    column indexes (A,B,C, ...)
    """
    self.drawarea.create_rectangle(self.__cellWidth,
                                   0,
                                   self.__cols[-1],
                                   self.__cellHeight,
                                   width=0,
                                   outline="grey",
                                   fill=color)

  def __drawRowBackground(self, color="grey"):
    """ <color:String> -> <None>

    Draws the background of the row indexes (1,2,3, ...)
    """
    self.drawarea.create_rectangle(0,
                                   self.__cellHeight,
                                   self.__cellWidth,
                                   self.__rows[-1],
                                   width=1,
                                   outline="grey",
                                   fill=color)

  def __drawColIndex(self):
    """<None> -> <None>

    Draws the column indexes (A,B,C, ...)
    """
    self.__colsTagIndex=[]
    column=1
    for i in xrange(1,len(self.__cols)):
      t=self.drawarea.create_text(\
        self.__cols[i-1]+(self.__cols[i]-self.__cols[i-1])*0.5,
        self.__cellHeight*0.5,
        text=intToString(column))
      self.__colsTagIndex.append(t)
      column+=1

  def __drawRowIndex(self):
    """<None> -> <None>

    Draws the row indexes (1,2,3, ...)
    """
    self.__rowsTagIndex=[]
    row=1
    for i in xrange(1,len(self.__rows)):
      t=self.drawarea.create_text(\
        self.__cellWidth*0.5,
        self.__rows[i-1]+(self.__rows[i]-self.__rows[i-1])*0.5,
        text=str(row))
      self.__rowsTagIndex.append(t)
      row+=1

  def __drawRows(self):
    """<None> -> <None>

    Draws the row lines. For performance reasons,
    each cell is not explicitely represented as a rectangle;
    instead, self.__rows horizontal and self.__cols vertical lines
    are drawn
    """
    self.__rowsTag=[]
    for i in xrange(0,len(self.__rows)):
      t=self.drawarea.create_line(0,
                                  self.__rows[i],
                                  self.__cols[-1],
                                  self.__rows[i],
                                  fill="grey")
      self.__rowsTag.append(t)

  def __drawColumns(self):
    """<None> -> <None>

    Draws the column lines.  For performance reasons 
    each cell is not explicitely represented as a rectangle;
    instead, self.__rows horizontal and self.__cols vertical lines
    are drawn
    """
    self.__colsTag=[]
    for i in xrange(0,len(self.__cols)):
      t=self.drawarea.create_line(self.__cols[i],
                                  0,
                                  self.__cols[i],
                                  self.__rows[-1],
                                  fill="grey")
      self.__colsTag.append(t)

  def __drawSelectedCell(self):
    """<None> -> <None>
    
    Redraw the selected cell (tick black rectangle) 
    in its new position. Called when the selected cell is 
    moved using moveSelectedCell
    """
    row=self.__selectedCell.row
    col=self.__selectedCell.col
    self.drawarea.delete(self.selectedCellTag)
    self.selectedCellTag = self.drawarea.create_rectangle(\
                             self.__cols[col],
                             self.__rows[row],
                             self.__cols[col+1],
                             self.__rows[row+1],
                             width=2, outline="black")

  def __drawOverCell(self):
    """<None> -> <None>

    Redraw the selected cell (thin blue rectangle) in its new position.
    Called when the over cell is moved using moveOverCell
    """ 
    row=self.__overCell.row
    col=self.__overCell.col
    self.drawarea.delete(self.overCellTag)
    self.overCellTag = self.drawarea.create_rectangle(\
                         self.__cols[col],
                         self.__rows[row],
                         self.__cols[col+1],
                         self.__rows[row+1],
                         width=1, outline="blue")

  def __getColBoundary(self, mouseX):
    cellCol = 0
    check=0
    for i in xrange(len(self.__cols)):
      if self.__cols[i]>mouseX:
        check=1
        if i==0:
          cellCol=0
        else:
          cellCol=i-1
        break
    if check:
      return (cellCol,cellCol+1)
    else:
      return (len(self.__cols)-2,len(self.__cols)-1)

  def __getRowBoundary(self, mouseY):
    cellRow = 0
    check=0
    for i in xrange(len(self.__rows)):
      if self.__rows[i]>mouseY:
        check=1
        if i==0:
          cellRow=0
        else:
          cellRow=i-1
        break
    if check:
      return (cellRow,cellRow+1)
    else:
      return (len(self.__rows)-2,len(self.__rows)-1)


  #**************************************************************************
  #
  # Next follows the interface to the gridview.  The following methods are
  # called from the statechart to express the dynamic behavior of the
  # spreadsheet.  For instance, one can move the selected cell (black tick
  # rectange) when the mouse is clicked using the moveSelectedCellTo method 
  #
  # *************************************************************************


  def getSelectedCell(self):
    """ <None> -> <CellCoord>

    Return the currently selected cell
    """
    return self.__selectedCell

  def getOverCell(self):
    """ <None> -> <CellCoord>

    Return the cell the mouse is currently over
    """
    return self.__overCell

  def getCellFormula(self, cell):
    """ <cell:CellCoord> -> <str>

    Return the formula expression (such as "=3+4") in a given cell in 
    String format. Return the empty string if no data is present in 
    the requested cell.
    """
    return self.subject.getFormula(CellCoord(cell.row+1,cell.col+1))

  def getCellValue(self, cell):
    """ <cell:CellCoord> -> <str>

    Returns the result of evaluating a formula in a given cell
    in String format. For example, formula "=3+4" returns "7".  Return
    the empty string if no data is present in the requested cell.
    """
    return self.subject.getValue(CellCoord(cell.row+1,cell.col+1))

  def setOverCellDisplay(self, string):
    """ <str> -> <None>

    Update the current edited formula display.
    Used for instance when the mouse is moved to
    display the formula expression of the cell under the mouse.
    """
    self.overCellData.set(string)

  def setColRowDisplay(self, string):
    """ <str> -> <None>

    Update the current row/column display.
    Used for instance when the mouse is moved to
    update the current row/column display of the cell
    under the mouse.
    """
    self.overCellString.set(string)


  #Other accessors methods to the class' private fields:

  def getHeight(self):
    return self.__height

  def getWidth(self):
    return self.__width
    
  def increaseTimeByOne(self):
    """<None> -> <None>

    Increases the time counter string by 1 second.
    Note that the time counter does not include hours
    (only "MM:SS"). The counter is reset to "00:00" after "59:59".
    Note: A more efficient implementation would keep tsec and tmin 
    as attributes rather than converting from the __timeData string
    every second.
    """
    t=self.__timeData.get()
    
    tsec=int(t[3:])
    
    if tsec<59:
      tsec+=1
      if tsec<10:
        self.__timeData.set(t[0:3]+"0"+str(tsec))
      else:
        self.__timeData.set(t[0:3]+str(tsec))
    else:
      tmin=int(t[0:2])
      if tmin<59:
        tmin+=1
        tsec=0
        if tmin<10:
          self.__timeData.set("0"+str(tmin)+":00")
        else:  
          self.__timeData.set(str(tmin)+":00")
      else:
        self.__timeData.set("00:00")
  
  def setTimeBlack(self):
    """<None> -> <None>

    Change the foreground/background of the time counter to white/black
    (reverse video)
    """
    self.__timeLabel.config(fg="white")
    self.__timeLabel.config(bg="black")

  def setTimeWhite(self):
    """<None> -> <None>

    Change the foreground/background of the time counter to black/white     
    """
    self.__timeLabel.config(fg="black")
    self.__timeLabel.config(bg="white")

  def moveSelectedCellTo(self, col, row):
    """<col:int> <row:int> -> <None>

    Move and redraw the selected cell
    to the coordinates (col,row)
    """
    if col<0:
      col=0
    elif col>=self.__width:
      col=self.__width-1
    if row<0:
      row=0
    elif row>=self.__height:
      row=self.__height-1

    self.__selectedCell.row=row
    self.__selectedCell.col=col
    self.__drawSelectedCell()

  def moveOverCellTo(self, col, row):
    """<col:int> <row:int> -> <None>

    Move and redraw the over cell
    to the coordinates (col,row)
    """
    if col<0:
      col=0
    elif col>=self.__width:
      col=self.__width-1
    if row<0:
      row=0
    elif row>=self.__height:
      row=self.__height-1

    self.__overCell.row=row
    self.__overCell.col=col
    self.__drawOverCell()

  def drawText(self,cellCoord,textLeft,textRight="",centered=0):
    """<cellCoord:CellCoord> <textLeft:str> 
       [<textRight="":str>] [<centered=0:int>] -> <None>

    Draws text at a given cell in the spreadsheet.  
    There is support for simple text editing as explained below.

    Each text that is drawn is divided into two parts: the part which is 
    to the ``left'' of the cursor and the part which is to the ``right'' 
    of the cursor.
    When the method ``drawCursor'' is called for a particular
    cell, it will draw the cursor in between what is to the ``left''
    and what is to the ``right''.  
    The text can also be centered horizontally in the cell.  
    To use this function to simply draw text ignoring cursor information
    (for instance, displaying values of the cells that are not being edited),
    simply omit to specify textRight. 
    Internally this will mean that the cursor is completely to the right 
    of the text, but if no cursor is drawn, this information will never 
    be used.  Note that the text that was in a cell before ``drawText'' 
    is called will be deleted by the new text. Only one text object is 
    allowed per cell.

    NOTE: The canvas text objects are not the data holder ! They must be
    kept IN SYNCH with the subject, which is the datastructure maintaining
    the cell formulas. The text objects can also contain data that is 
    currently being edited and not yet submitted to the subject 
    (see acceptData).  
    Note one can move the cursor with two calls to drawText:
      1)  drawText(cell, "abc", "def")  ==> abc|def
      2)  drawText(cell, "abcd", "ef")  ==> abcd|ef
    """
    col=cellCoord.col
    row=cellCoord.row
    if self.textOnCanvas.has_key((cellCoord.col,cellCoord.row)):
      tag=self.textOnCanvas[(cellCoord.col,cellCoord.row)]
      self.drawarea.delete(tag[0])
      self.drawarea.delete(tag[1])
    if not centered:
      deltaX=self.__cols[col]+1
    else:
      deltaX=self.__cols[col]+(self.__cols[col+1]-self.__cols[col])*0.5
    deltaY=self.__rows[row]+(self.__rows[row+1]-self.__rows[row])*0.5
    if not centered:
      tagLeft=self.drawarea.create_text(deltaX,
                                        deltaY,
                                        text=textLeft,
                                        justify=CENTER,
                                        anchor="w")
    else:
      tagLeft=self.drawarea.create_text(deltaX,
                                        deltaY,
                                        text=textLeft,
                                        justify=CENTER)
    textLeftBox=self.drawarea.bbox(tagLeft)
    tagRight=self.drawarea.create_text(textLeftBox[2],
                                       (textLeftBox[3]+textLeftBox[1])/2,
                                       text=textRight,
                                       justify=LEFT,
                                       anchor="w")

    self.textOnCanvas[(col,row)]=(tagLeft,tagRight)


  def drawCursor(self,cellCoord):
    """<cellCoord:CellCoord> -> <None>

    Animate a cursor with respect to the text that is in cellCoord 
    (see drawText).
    Only one cursor is allowed at a given time in the spreadsheet.
    If one wants to create a cursor in another cell, ``deleteCursor''
    must be called first.  If drawCursor is called two times in a row
    it will become impossible to delete one of the cursors.
    The pattern ``deleteCursor'' & ``drawCursor'' needs to be used
    each time the current edited text is modified with drawText to move
    the cursor accordingly.
    """
    textTagLeft=self.textOnCanvas[(cellCoord.col,cellCoord.row)][0]
    textBox=self.drawarea.bbox(textTagLeft)

    X1=textBox[2]
    Y1=textBox[1]
    X2=textBox[2]
    Y2=textBox[3]

    tag=self.drawarea.create_line(X1,Y1,X2,Y2)
    deleteEvent=self.root.after(500,self.drawarea.delete,tag)
    creationEvent=self.root.after(1000,self.drawCursor,cellCoord)
    self.textOnCanvas["cursor"]=(tag,deleteEvent,creationEvent,cellCoord)

  def deleteCursor(self):
    """<None> -> <None>

    Delete the current animated cursor.
    Is used in conjunction with ``drawCursor'' and
    ``drawText'' to move the cursor within the same edited cell.
    """
    if self.textOnCanvas.has_key("cursor"):
      cursorInfo=self.textOnCanvas["cursor"]
      self.drawarea.delete(cursorInfo[0])
      self.root.after_cancel(cursorInfo[1])
      self.root.after_cancel(cursorInfo[2])
      del self.textOnCanvas["cursor"]

  # Helper methods to check if a keypress is alphachar, nonalphachar, etc
  # to determine the type of character typed in.
  # All is... methods return Boolean True (1) or False (0)

  def isAlphaChar(self, keysym):
   """<char> -> <Boolean>

   <AlphaChar> in the range a - z, A - Z
   """
   return ( (len(keysym) == 1) and
            (('a' <= keysym <= 'z') or
             ('A' <= keysym <= 'Z')) )

  def isNonAlphaChar(self, keysym):
   """<char> -> <Boolean>

   <NonAlphaChar> in { /~!@#$%^&*()-_+;:'",<>?[]}
   """
   nonAlphaChars = ['slash',         # /
                    'asciitilde',    # ~
                    'exclam',        # !
                    'at',            # @
                    'numbersign',    # #
                    'dollar',        # $
                    'percent',       # %
                    'asciicircum',   # ^
                    'ampersand',     # &
                    'asterisk',      # *
                    'parenleft',     # (
                    'parenright',    # )
                    'minus',         # -
                    'underscore',    # _
                    'plus',          # +
                    'semicolon',     # ;
                    'colon',         # :
                    'apostrophe',    # '
                    'quotedbl',      # "
                    'comma',         # ,
                    'less',          # <
                    'greater',       # >
                    'question',      # ?
                    'bracketleft',   # [
                    'bracketright']  # ]
   return keysym in nonAlphaChars

  def isDigitChar(self, keysym):
   """<char> -> <Boolean>

   <DigitChar> in the range 0 - 9
   """
   return ('0' <= keysym <=  '9')

  def isPeriod(self, keysym):
   """<char> -> <Boolean>

   <period> = '.'
   """
   return (keysym == 'period')

  def isEquals(self, keysym):
   """<char> -> <Boolean>

   <equals> = '='
   """
   return (keysym == 'equal')

  def cellCoords(self, mouseX, mouseY):
    """<mouseX:int> <mouseY:int> -> <CellCoord>

    Utility method to convert mouse position (event.x, event.y)
    into cell coordinates (row, col)
    """
    cellRow=self.__getRowBoundary(mouseY)[0]
    cellCol=self.__getColBoundary(mouseX)[0]
    return CellCoord(cellRow, cellCol)

  def acceptData(self,cell,data):
      """<cell:CellCoord> <data:str> -> <Boolean>

      Commit data with respect to a particular cell in the subject.  
      This method either succeeds or fails depending on whether there
      is a parse error when processing the formula.  
      The data is NOT commited if the method  returns false, it is 
      commited otherwise.  A message is printed to stout if there is 
      a parse error.  
      NOTE: Before, a tkMessageBox was used to display the error, 
      but creating such an object from SVM freezes the system 
      (GOD knows why: SVM is using threads and Tkinter does strange
      things with threads).
      """
      try:
        if data=="":
          return True 
        if len(data)>0:
          if data[0]!='=':
            try:
              data=float(data)
            except Exception,e:
              pass
            if type(data)==types.FloatType:
              data="="+str(data)
        self.subject.setState(CellCoord(cell.row+1,cell.col+1),data)
      # HV: the exceptions below don't seem to be caught
      except FormulaParseError:
        #tkMessageBox.showwarning("Formula Error",
	# "A parse error was found in the formula")
        print "A parse error was found in the formula"
        return False 
      except FunctionNameError,e:
        #tkMessageBox.showwarning("Formula Error",
	# "Parse Error!"+str(e)+" does not exists")
        print "Parse Error!",str(e)+" does not exists"        
        return False 
      return True 

  # The two following functions were supposed to be called from the statechart,
  # but due to the fact that SVM uses threads, creating dialogs such as
  # asksaveasfilename hangs the interpreter.  Thus, this behavior was
  # left in this class.  (Note that button press events are bound to
  #  these 2 functions).

  def printView(self):
    printFilename = \
     tkFileDialog.asksaveasfilename(filetypes=[("Postscript files", "*.ps")])
    bbox=self.drawarea.bbox(ALL)
    self.drawarea.postscript(file=printFilename,
                             width=bbox[2]+10, 
			     height=bbox[3]+10)

  def wmQuit(self):
    self.dynamicGridView.stop()
    print self.__title + " was closed (by Window Manager)"
    self.root.destroy()
    self.subject.destroy()

# The SpreadsheetGridView_Dynamic class ``bridges'' the 
# + non-behaviour information encoded in the class SSGridview and 
# + the behaviour information modelled in the statechart
#   (encoded in the class SSGridViewStateChart generated from the
#    statechart).  
# GUI events are bound to the methods on...() of the 
# SpreadsheetGridView_Dynamic class.
# These methods in turns trigger an event in
# the statechart.  Then, the statechart reacts to the event
# by calling (back) appropriate functions of the SSGridView class
# to execute the correct behavior.

class SpreadsheetGridView_Dynamic:

  def __init__(self,staticGV):
    self.statechart=SSGridViewStateChart()
    self.statechart.initModel()
    self.statechart.event("start",staticGV)

  def onDrawareaMouseMove(self, event):
    self.statechart.event("mouseMove",event)

  def onDrawareaMouse1Click(self, event):
    self.statechart.event("mouseClick",event)

  def onDrawareaLeft(self, event):    
    self.statechart.event("leftKey")

  def onDrawareaRight(self, event):
    self.statechart.event("rightKey")

  def onDrawareaUp(self, event):
    self.statechart.event("upKey")

  def onDrawareaDown(self, event):
    self.statechart.event("downKey")

  def onDrawareaKeyPressed(self, event):
    self.statechart.event("keypress",event)

  def onDrawareaDoubleClick(self, event):
    self.statechart.event("doubleClick",event)
  
  def printView(self):
    self.statechart.event("print")
  
  def stop(self):
    self.statechart.event("quit")
  

